#!/usr/bin/env python
# -*- coding: UTF-8 -*-

## This program is a free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation, version 3 of the License.
##
## You can get a copy of GNU General Public License along this program
## But you can always get it from http://www.gnu.org/licenses/gpl.txt
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
## GNU General Public License for more details.
##
## Author: andreyefgs
## http://forum.ru-board.com/topic.cgi?forum=93&topic=3913&glp
##
## Many thanks to:
## ilius (Saeed Rasooli)
## https://github.com/ilius/pyglossary

import time
start = time.time()
import sys
import os
import re
from os.path import isfile
import struct
import codecs
from functools import cmp_to_key
import subprocess

def runDictzip(filename):
    dictzipCmd = '/usr/bin/dictzip'
    if not os.path.isfile(dictzipCmd):
        return False
    if filename[-4:]=='.ifo':
        filename = filename[:-4]
    (out, err) = subprocess.Popen(
        [dictzipCmd, filename + '.dict'],
        stdout=subprocess.PIPE
    ).communicate()


def utf8len(s):
	return len(s.encode('utf-8'))

def trimlist(l):
	"""
		['dog ', '', ' pig']
		['dog', 'pig'] 
	"""
	tlist = [item.strip() for item in l]
	flist = list(filter(None, tlist))
	return flist

def isAsciiUpper(c):
	"""
		imitate ISUPPER macro of glib library gstrfuncs.c file
	"""
	return ord(c) >= ord('A') and ord(c) <= ord('Z')

def asciiLower(c):
	"""
		imitate TOLOWER macro of glib library gstrfuncs.c file

		This function converts upper case Latin letters to corresponding lower case letters,
		other chars are not changed.

		c must be non-Unicode string of length 1.
		You may apply this function to individual bytes of non-Unicode string.
		The following encodings are allowed: single byte encoding like koi8-r, cp1250, cp1251, cp1252, etc,
		and utf-8 encoding.

		Attention! Python Standard Library provides str.lower() method.
		It is not a correct replacement for this function.
		For non-unicode string str.lower() is locale dependent, it not only converts Latin
		letters to lower case, but also locale specific letters will be converted.
	"""
	if isAsciiUpper(c):
		return chr((ord(c) - ord('A')) + ord('a'))
	else:
		return c


def asciiStrCaseCmp(s1, s2):
	"""
		imitate g_ascii_strcasecmp function of glib library gstrfuncs.c file
	"""
	commonLen = min(len(s1), len(s2))
	for i in range(commonLen):
		c1 = ord(asciiLower(s1[i]))
		c2 = ord(asciiLower(s2[i]))
		if c1 != c2:
			return c1 - c2
	return len(s1) - len(s2)

def strCmp(s1, s2):
	"""
		imitate strcmp of standard C library

		Attention! You may have a temptation to replace this function with built-in cmp() function.
		Hold on! Most probably these two function behave identically now, but cmp does not
		document how it compares strings. There is no guaranty it will not be changed in future.
		Since we need predictable sorting order in StarDict dictionary, we need to preserve
		this function despite the fact there are other ways to implement it.
	"""
	commonLen = min(len(s1), len(s2))
	for i in range(commonLen):
		c1 = ord(s1[i])
		c2 = ord(s2[i])
		if c1 != c2:
			return c1 - c2
	return len(s1) - len(s2)

def stardictStrCmp(s1, s2):
	"""
		use this function to sort index items in StarDict dictionary
		s1 and s2 must be utf-8 encoded strings
	"""
	a = asciiStrCaseCmp(s1, s2)
	if a == 0:
		return strCmp(s1, s2)
	else:
		return a

 
def scmp(a,b):
	return stardictStrCmp(a[0],b[0])



def cli_progress_bar(title, pos, end_val, bar_length=20):
	percent = float(pos) / end_val
	hashes = '#' * int(round(percent * bar_length))
	spaces = ' ' * (bar_length - len(hashes))
	sys.stdout.write("\r" + title + ": [{0}] {1}%".format(hashes + spaces, int(round(percent * 100))))
	sys.stdout.flush()

class StarDictReader:
	def __init__(self, path):
		self.filePath = path
		self.fileBasePath = os.path.splitext(path)[0]
		self.ifoData = {}
		self.synData = {}
		self.indexData = []

	def run(self):
		self.readIfoFile()
		self.readIdxFile()
		self.readSynFile()
		self.readDictFile()
		print('Execution time: {0:0.1f} seconds'.format(time.time() - start))

	def readIfoFile(self):
		try:
			ifo = codecs.open(self.filePath, 'r', encoding='utf8')
		except IOError:
			print("\nCould not open file: " + self.filePath)	
			sys.exit()
	
		for line in ifo.readlines():
			m = re.match(u'^(bookname|wordcount|idxfilesize|sametypesequence|description|author|synwordcount)=(.*)$', line)
			if m:
				self.ifoData[m.group(1)] = m.group(2).strip()
	
		iSifovalid = False
		if "bookname" in self.ifoData and len(self.ifoData["bookname"]) > 0 \
		and "wordcount" in self.ifoData and re.match(u'^[1-9]\d*$', self.ifoData["wordcount"]) \
		and "idxfilesize" in self.ifoData and re.match(u'^[1-9]\d*$', self.ifoData["idxfilesize"]) \
		and "sametypesequence" in self.ifoData and re.match(u'(m|l|g|t|x|y|k|w|h|r|W|P|X)$', self.ifoData["sametypesequence"]) \
		and (("synwordcount" not in self.ifoData) or ("synwordcount" in self.ifoData) and re.match(u'^[1-9]\d*$', self.ifoData["synwordcount"])):
			iSifovalid = True
	
	
		if not iSifovalid:
			print("\nInvalid ifo file")
			sys.exit()
			
		if self.ifoData["sametypesequence"] != 'h':
			print("Definition format {} is not supported".format(self.ifoData["sametypesequence"]))
			sys.exit()
			
	
	
	def readIdxFile(self):
		if isfile(self.fileBasePath + '.idx.gz'):
			import gzip
			try:
				idx = gzip.open(self.fileBasePath + '.idx.gz')
			except IOError:
				print("\nCould not open file: " + self.fileBasePath + '.idx.gz')	
				sys.exit()
		else:
			try:
				idx = open(self.fileBasePath +'.idx', 'rb')
			except IOError:
				print("\nCould not open file: " + self.fileBasePath + '.idx')	
				sys.exit()
		idxStr = idx.read()
		i = 0
		while i < len(idxStr):
			beg = i
			i = idxStr.find(b'\x00', beg)
			if i < 0:
				print("Index file is corrupted.")
				sys.exit()
			word = idxStr[beg:i]
			i += 1
			if i + 8 > len(idxStr):
				print("Index file is corrupted.")
				sys.exit()
			offset = struct.unpack(">I", idxStr[i:i+4])[0]
			i += 4
			size = struct.unpack(">I", idxStr[i:i+4])[0]
			i += 4
			self.indexData.append([word.decode('utf-8'), offset, size])
	
	
	def readSynFile(self):

		indexCount = len(self.indexData)
	
		if not isfile(self.fileBasePath + '.syn'):
			return
	
		try:
			synStr = open(self.fileBasePath + '.syn', 'rb').read()
		except IOError:
			print("\nCould not open file: " + self.fileBasePath + '.syn')	
			sys.exit()
	
		synStrLen = len(synStr)
		i = 0
		while i < synStrLen:
			beg = i
			i = synStr.find(b'\x00', beg)
			if i < 0:
				brint("Synonym file is corrupted.")
				sys.exit()
			word = synStr[beg:i]
			i += 1
			if i + 4 > len(synStr):
				print("Synonym file is corrupted.")
				sys.exit()
			index = struct.unpack(">I", synStr[i:i+4])[0]
			i += 4
			if index >= indexCount:
				print("Corrupted synonym file. Word \"{0}\" references invalid item.".format(word))
				continue
			
			try:
				self.synData[index].append(word.decode('utf-8'))
			except KeyError:
				self.synData[index] = [word.decode('utf-8')]
	
	
	def readDictFile(self):
		if isfile(self.fileBasePath + '.dict.dz'):
			import gzip
			try:
				dictFd = gzip.open(self.fileBasePath + '.dict.dz')
			except IOError:
				print("\nCould not open file: " + self.fileBasePath + '.dict.dz')	
				sys.exit()
		else:
			try:
				dictFd = open(self.fileBasePath + '.dict', 'rb')
			except IOError:
				print("\nCould not open file: " + self.fileBasePath + '.dict')	
				sys.exit()
	
		try:
			gls = codecs.open(self.fileBasePath + ".gls", 'w', encoding='utf8')
		except IOError:
			print("\nCould not open file: " + self.fileBasePath + '.gls')	
			sys.exit()
	
		end_val = len(self.indexData)
		artcount = 0
	
		IfoStr = "\n#bookname=" + self.ifoData["bookname"] + "\n"
		if "description" in self.ifoData:
			IfoStr += "#description=" + self.ifoData["description"] + "\n"
		IfoStr += "#sametypesequence=h\n\n"
	
		gls.write(IfoStr)
	
		for index, (word, defiOffset, defiSize) in enumerate(self.indexData):
	
			artcount += 1;
			if artcount % 20 == 0:
				cli_progress_bar("Writing", artcount, end_val, bar_length=40)
	
	
			if not word:
				continue
	
			dictFd.seek(defiOffset)
			if dictFd.tell() != defiOffset:
				print("Unable to read definition for word \"{0}\"".format(word))
				continue
	
			data = dictFd.read(defiSize)
	
			if len(data) != defiSize:
				print("Unable to read definition for word \"{0}\"".format(word))
				continue
	
			gls.write(word)
			list = self.synData.get(index, [])
			if len(list) > 0:
				gls.write('|' + "|".join(list))
			gls.write("\n")
			gls.write(data.decode('utf-8'))
			gls.write("\n\n")
	
		cli_progress_bar("Writing", end_val, end_val, bar_length=40)
		print("")
		dictFd.close()
		gls.close()


class StarDictWriter:
	def __init__(self, path):
		self.filePath = path
		self.fileBasePath = os.path.splitext(path)[0]
		self.glos = []
		self.header = {}
		self.alternates = []

	def run(self):
		self.readHeader()
		self.readArticles()
		self.writeGeneral()
		if len(self.alternates) > 0:
			self.writeSynFile()
		self.writeIfoFile()
		runDictzip(self.fileBasePath)
		print("\nwordcount: " + str(len(self.glos)))
		if (len(self.alternates) > 0):
			print("synwordcount: " + str(len(self.alternates)))
		print('Execution time: {0:0.1f} seconds'.format(time.time() - start))

	def readHeader(self):
		linecount = 0
		try:
			src = codecs.open(self.filePath, 'r', encoding='utf8')
		except IOError:
			print("\nCould not open file: " + self.filePath)	
			sys.exit()
		for line in src.readlines():
			line = line.rstrip()
			if re.match(u'^\ufeff.*$', line):
				line = re.sub(u'^\ufeff', '', line)
			linecount += 1
			if linecount > 20:
				break
			if re.match(u'^[^#].*$', line) and len(line.strip()) > 0:
				break
			m = re.match(u'^###\x20?(Glossary title|Author|Description)\x20*[:\?]\x20*(.+)$', line)
			if m:
				if len(m.group(2).strip()) > 0:
					self.header[m.group(1)] = m.group(2).strip()
			m = re.match(u'^#(bookname|author|description)\x20*[=]\x20*(.+)$', line)
			if m:
				if len(m.group(2).strip()) > 0:
					self.header[m.group(1)] = m.group(2).strip()
		src.close()


	def readArticles(self):
		words = ""
		defis = ""
		notemptylinecount = 0
		ignore_header_type_records = 0
		linecount = 0
		try:
			src = codecs.open(self.filePath, 'r', encoding='utf8')
		except IOError:
			print("\nCould not open file: " + self.filePath)	
			sys.exit()
		statinfo = os.stat(self.filePath)
		end_val = statinfo.st_size
	
	
		for line in iter(src.readline, ''):
			linecount += 1
	
			if linecount % 20 == 0:
				pos = src.tell()
				cli_progress_bar("Reading", pos, end_val, bar_length=40)
			
			line = line.rstrip()
			if re.match(u'^\ufeff.*$', line):
				line = re.sub(u'^\ufeff', '', line)
			if len(line.strip()) > 0:
				if ignore_header_type_records == 0:
					if re.match(u'^#{1,3}\x20?(Glossary title|Author|Description|Source language|Source alphabet|Target language|Target alphabet|Icon|Browsing enabled|Type of glossary|Case sensitive words|Glossary section|bookname|author|description|wordcount|synwordcount|idxfilesize|idxoffsetbits|email|website|date|sametypesequence|stripmethod).*$', line):
						continue
				if notemptylinecount == 0:
					words = line.strip()
				elif notemptylinecount == 1:
					defis = line
	
					synonymlist = []
					word = ""
					if '|' in words:
						list = trimlist(words.split("|"))
						if len(list) == 1:
							word = list[0]
						elif len(list) > 1:
							word = list[0]
							for synonym in list[1:]:
								if utf8len(synonym) > 255:
									print("\nLine {}, the headword is too long".format(linecount))
									sys.exit()
								synonymlist.append(synonym)
						else:
							print("\nLine {}, bad word!!!".format(linecount))
							sys.exit()
					else:
						word = words
	
					if utf8len(word) > 255:
						print("\nLine {}, the headword is too long".format(linecount))
						sys.exit()
	
					self.glos.append([word, defis, synonymlist])
					
					words = ""
					defis = ""
				else:
					print("\nError, expected an empty line after definition: " + str(linecount))
					sys.exit()
				
				ignore_header_type_records = 1
				notemptylinecount += 1
		
			else:
		
				if words != "":
					print("\nError, found an empty line where not expected: " + str(linecount))
					sys.exit()
		
				notemptylinecount = 0
		if (words != ""):
			print("\nError, orphaned term or definition: " + str(linecount))
			sys.exit()
		cli_progress_bar("Reading", end_val, end_val, bar_length=40)
		src.close()
	
	
	def writeSynFile(self):
		try:
			syn = codecs.open(self.fileBasePath + ".syn", 'wb')
		except IOError:
			print("\nCould not open file: " + self.fileBasePath + ".syn")	
			sys.exit()
		if len(self.alternates) > 0:
			sys.stdout.write("\nWriting synonyms...")
			self.alternates.sort(key=cmp_to_key(scmp))
			synStr = ''
			for item in self.alternates:
				synStr = item[0] + '\x00'
				syn.write(synStr.encode('utf-8'))
				syn.write(struct.pack(">I", item[1]))
			del synStr
			sys.stdout.write("\rWriting synonyms...Ok\n")
		syn.close()
	
	
	def writeGeneral(self):
		try:
			dic = codecs.open(self.fileBasePath + ".dict", 'w', encoding='utf8')
		except IOError:
			print("\nCould not open file: " + self.fileBasePath + ".dict")	
			sys.exit()
	
		try:
			ind = codecs.open(self.fileBasePath + ".idx", 'wb')
		except IOError:
			print("\nCould not open file: " + self.fileBasePath + ".idx")	
			sys.exit()
	
		artcount = 0
		dictMark = 0
	
		sys.stdout.write("\nSorting...")
		self.glos.sort(key=cmp_to_key(scmp))
		end_val = len(self.glos)
		sys.stdout.write("\rSorting...Ok\n")
	
		for i, art in enumerate(self.glos):
			artcount += 1;
			if artcount % 10 == 0:
				cli_progress_bar("Writing", artcount, end_val, bar_length=40)
	
			dictBlock = ''
			defiFormat = ''
		
			
			dictBlock += defiFormat + art[1]
			dic.write(dictBlock)
		
			dataLen = utf8len(dictBlock)
	
			for altWord in art[2]:
				self.alternates.append((altWord, i))
			
			idxBlock = art[0] + '\x00'
			ind.write(idxBlock.encode('utf-8'))
			ind.write(struct.pack(">I", dictMark))
			ind.write(struct.pack(">I", dataLen))
		
			dictMark += dataLen
		cli_progress_bar("Writing", end_val, end_val, bar_length=40)		
		dic.close()
		ind.close()
	
	def writeIfoFile(self):
		try:
			ifo = codecs.open(self.fileBasePath + ".ifo", 'w', encoding='utf8')
		except IOError:
			print("\nCould not open file: " + self.fileBasePath + ".ifo")	
			sys.exit()
	
		statinfo = os.stat(self.fileBasePath + ".idx")
		author = ""
		description = ""
	
		if "Glossary title" in self.header:
			dicname = self.header["Glossary title"]
		else:
			if "bookname" in self.header:
				dicname = self.header["bookname"]
		
		if "Description" in self.header:
			description = self.header["Description"]
		else:
			if "description" in self.header:
				description = self.header["description"]
		
		if "Author" in self.header:
			author = self.header["Author"]
		else:
			if "author" in self.header:
				author = self.header["author"]
		
		ifoStr = "StarDict's dict ifo file\n" \
			+ "version=3.0.0\n" \
			+ "bookname={0}\n".format(dicname) \
			+ "wordcount={0}\n".format(len(self.glos)) \
			+ "idxfilesize={0}\n".format(statinfo.st_size) \
			+ "sametypesequence={0}\n".format("h") \
			+ "description={0}\n".format(description) \
			+ "author={0}\n".format(author)
	
		if len(self.alternates) > 0:
			ifoStr += "synwordcount={0}\n".format(len(self.alternates))
	
		ifo.write(ifoStr)
		ifo.close()



if len(sys.argv) != 2:
	print("\nInvalid command line")
	sys.exit()
else:
	if re.match(u'^\.gls$', os.path.splitext(sys.argv[1])[1], re.IGNORECASE):
		writer = StarDictWriter(sys.argv[1])
		writer.run()
	elif os.path.splitext(sys.argv[1])[1] == '.ifo':  
		reader = StarDictReader(sys.argv[1])
		reader.run()
	else:
		print("\nInvalid command line")
		sys.exit()
